#!/usr/bin/env python3

"""Browse autopkgtest results"""

import os
import sqlite3
import json
import configparser
import urllib.parse
import re
import distro_info
from collections import OrderedDict

from wsgiref.handlers import CGIHandler
import flask

app = flask.Flask('browse')
db_con = None
swift_container_url = None

UDI = distro_info.UbuntuDistroInfo()
ALL_UBUNTU_RELEASES = UDI.all
SUPPORTED_UBUNTU_RELEASES = UDI.supported()


def init_config():
    global db_con, swift_container_url

    cp = configparser.ConfigParser()
    cp.read(os.path.expanduser('~ubuntu/autopkgtest-cloud.conf'))

    db_con = sqlite3.connect('file:%s?mode=ro' % cp['web']['database_ro'], uri=True)
    try:
        url = cp['web']['ExternalURL']
    except KeyError:
        url = cp['web']['SwiftURL']
    swift_container_url = os.path.join(url, 'autopkgtest-%s')


def get_test_id(release, arch, src):
    c = db_con.cursor()
    c.execute('SELECT id FROM test WHERE release=? AND arch=? AND package=?',
              (release, arch, src))
    try:
        return c.fetchone()[0]
    except TypeError:
        return None


def render(template, code=200, **kwargs):
    # sort the values passed in, so that releases are in the right order
    try:
        release_arches = OrderedDict()
        for k in sorted(kwargs['release_arches'], key=ALL_UBUNTU_RELEASES.index):
            release_arches[k] = kwargs['release_arches'][k]
        kwargs['release_arches'] = release_arches
    except KeyError:
        pass
    try:
        kwargs['releases'] = sorted(kwargs['releases'], key=ALL_UBUNTU_RELEASES.index)
    except KeyError:
        pass
    return (flask.render_template(template,
                                  base_url=flask.url_for('index_root'),
                                  static_url=flask.url_for('static', filename='/'),
                                  **kwargs), code)


def human_date(run_id):
    return re.sub(r'(\d\d\d\d)(\d\d)(\d\d)_(\d\d)(\d\d)(\d\d).*',
                  r'\1-\2-\3 \4:\5:\6 UTC',
                  run_id)


def human_sec(secs):
    return '%ih %02im %02is' % (secs // 3600, (secs % 3600) // 60, secs % 60)


def human_exitcode(code):
    if code in (0, 2):
        return 'pass'
    if code in (4, 6, 12):
        return 'fail'
    if code in (8,):
        return 'neutral'
    if code == 99:
        return 'blacklisted'
    if code == 16:
        return 'tmpfail'
    if code == 20:
        return 'error'
    return 'code %i' % code


def srchash(src):
    if src.startswith('lib'):
        return src[:4]
    else:
        return src[0]


def get_release_arches():
    '''Determine available releases and architectures

    Return release → [arch] dict.
    '''
    release_arches = {}
    releases = []
    for row in db_con.execute('SELECT DISTINCT release from test'):
        if row[0] in SUPPORTED_UBUNTU_RELEASES:
            releases.append(row[0])
    for r in releases:
        for row in db_con.execute('SELECT DISTINCT arch from test WHERE release=?', (r,)):
            release_arches.setdefault(r, []).append(row[0])
    return release_arches


def get_queue_info():
    '''Return information about queued tests

    Return (releases, arches, context -> release -> arch -> (queue_size, [requests])).
    '''

    with open('/var/lib/cache-amqp/queued.json', 'r') as json_file:
        queue_info_j = json.load(json_file)

        releases = queue_info_j["releases"]
        arches = queue_info_j["arches"]
        queues = queue_info_j["queues"]

        ctx = {}

        for context in queues:
            for release in queues[context]:
                for arch in queues[context][release]:
                    requests = queues[context][release][arch]["requests"]
                    size = queues[context][release][arch]["size"]
                    ctx.setdefault(context, {}).setdefault(release, {})[arch] = (size, requests)

        return (releases, arches, ctx)


def get_source_versions(db, release):
    '''Return srcpkg → current_version mapping for given release'''

    srcs = {}
    for (pkg, ver) in db.execute('SELECT package, version '
                                 'FROM current_version '
                                 'WHERE release = ?', (release, )):
        srcs[pkg] = ver
    return srcs


def success_count_for_release_and_arch(db, release, arch, src_versions):
    '''Return number of packages with tests that pass'''

    count = 0

    # we consider the latest successful run of a package's current version with
    # triggers for current packages; if none exist (i. e. tests generally fail,
    # but succeeded for a trigger that is not published), don't count it as
    # success
    cur_pkg = None
    for (pkg, triggers, code) in db.execute(
            'SELECT test.package, triggers, exitcode '
            'FROM test, result, current_version '
            'WHERE test.id == result.test_id AND test.release=? AND arch=? '
            '  AND test.package = current_version.package '
            '  AND test.release = current_version.release '
            '  AND result.version = current_version.version '
            '  AND (exitcode = 0 OR exitcode = 2 OR exitcode = 8) '
            'ORDER BY test.package, run_id DESC', (release, arch)):
        # start of a new package block?
        if pkg != cur_pkg:
            # logging.debug('new package start: %s [%s] %i', pkg, triggers, code)
            cur_pkg = pkg
            found_valid = False
        elif found_valid:
            # we found a valid success, skip over the older test results of
            # that package
            # logging.debug('ignored older result: %s [%s] %i', pkg, triggers, code)
            continue
        # logging.debug('considered result: %s [%s] %i', pkg, triggers, code)
        # weed out non-current triggers
        for trigger in triggers.split():
            src, ver = trigger.split('/')
            # it can happen that src_versions does not have a trigger source
            # pacakge if that trigger source got removed in the final release
            if src_versions.get(src) != ver:
                # logging.debug('%s/%s/%s: ignoring non-current trigger %s', release, arch, pkg, trigger)
                continue
        # if we arrive here, we have a passing test result for pkg with current
        # triggers, so count it
        # logging.debug('counting result: %s [%s] %i', pkg, triggers, code)
        count += 1
        # ... and ignore older results for that package
        found_valid = True

    return count


@app.route('/')
def index_root():
    letters = list('abcdefghijklmnopqrstuvwxyz')
    indexes = letters + ['lib' + l for l in letters]
    indexes.sort()

    recent = []
    for row in db_con.execute('SELECT exitcode, package, release, arch, triggers '
                              'FROM result, test '
                              'WHERE test.id == result.test_id '
                              'ORDER BY run_id DESC '
                              'LIMIT 15'):
        hc = human_exitcode(row[0])
        res = hc if 'code' not in hc else 'fail'
        recent.append((res, row[1], row[2], row[3], row[4]))

    return render('browse-home.html',
                  indexes=indexes,
                  recent_runs=recent)


# backwards-compatible path with debci that specifies the source hash
@app.route('/packages/<_>/<package>')
@app.route('/packages/<package>')
def package_overview(package, _=None):
    results = {}
    arches = set()
    for row in db_con.execute('SELECT MAX(run_id), exitcode, release, arch '
                              'FROM test, result '
                              'WHERE package = ? AND test.id = result.test_id '
                              'GROUP BY release, arch', (package,)):
        arches.add(row[3])
        results.setdefault(row[2], {})[row[3]] = human_exitcode(row[1])

    return render('browse-package.html',
                  package=package,
                  releases=[release for release in results.keys() if release in SUPPORTED_UBUNTU_RELEASES],
                  arches=sorted(arches),
                  results=results,
                  title_suffix='- %s' % package)


# backwards-compatible path with debci that specifies the source hash
@app.route('/packages/<_>/<package>/<release>/<arch>')
@app.route('/packages/<package>/<release>/<arch>')
def package_release_arch(package, release, arch, _=None):
    test_id = get_test_id(release, arch, package)
    if test_id is None:
        return render('browse-error.html', error='Package does not exist', code=404)

    seen = set()
    results = []
    for row in db_con.execute('SELECT run_id, version, triggers, duration, exitcode, requester FROM result '
                              'WHERE test_id=? '
                              'ORDER BY run_id DESC', (test_id,)):
        requester = row[5] if row[5] else '-'
        code = human_exitcode(row[4])
        identifier = (row[1], row[2])  # Version + triggers uniquely identifies this result
        show_retry = code != 'pass' and identifier not in seen
        seen.add(identifier)
        url = os.path.join(swift_container_url % release, release, arch, srchash(package), package, row[0])
        results.append((row[1], row[2], human_date(row[0]), human_sec(row[3]), requester, code, url, show_retry))

    return render('browse-results.html',
                  package=package,
                  release=release,
                  arch=arch,
                  package_results=results,
                  title_suffix='- %s/%s/%s' % (package, release, arch))


@app.route('/running')
def running():
    (releases, arches, queue_info) = get_queue_info()
    queue_lengths = {}
    for c in queue_info:
        for r in releases:
            for a in arches:
                (queue_length, queue_items) = queue_info.get(c, {}).get(r, {}).get(a, (0, []))
                queue_lengths.setdefault(c, {}).setdefault(r, {})[a] = queue_length

    try:
        with open('/run/amqp-status-collector/running.json') as f:
            # package -> runhash -> release -> arch -> (params, duration, logtail)
            running_info = json.load(f)
    except FileNotFoundError:
        running_info = {}

    return render('browse-running.html',
                  contexts=queue_info.keys(),
                  releases=releases,
                  arches=arches,
                  queue_info=queue_info,
                  queue_lengths=queue_lengths,
                  running=running_info)

@app.route('/queue_size.json')
def queuesize_json():
    out = {}
    queue_info = get_queue_info()[2]
    # Strip the number of queue items, this is just their contents
    for context in queue_info:
        for release in queue_info[context]:
            for arch in queue_info[context][release]:
                (queue_size, queue_items) = queue_info[context][release][arch]
                queue_info[context][release][arch] = len(queue_items)
    return json.dumps(queue_info, indent=2)

@app.route('/queues.json')
def queues_json():
    out = {}
    queue_info = get_queue_info()[2]
    # Strip the number of queue items, this is just their contents
    for context in queue_info:
        for release in queue_info[context]:
            for arch in queue_info[context][release]:
                (queue_size, queue_items) = queue_info[context][release][arch]
                queue_info[context][release][arch] = queue_items
    return json.dumps(queue_info, indent=2)


@app.route('/testlist')
def testlist():
    # geneate index → [(package, version)] map
    indexed_pkgs = {}
    for row in db_con.execute('SELECT package, MAX(version) '
                              'FROM test, result '
                              'WHERE id == test_id '
                              'GROUP BY package '
                              'ORDER BY package'):
        # strip off epoch
        v = row[1][row[1].find(':') + 1:]
        indexed_pkgs.setdefault(srchash(row[0]), []).append((row[0], v))
    return render('browse-testlist.html',
                  indexed_pkgs=indexed_pkgs)


@app.route('/statistics')
def statistics():
    release_arches = get_release_arches()

    # ensure we don't run into KeyErrors in the jinja template
    data = {}
    for release in release_arches:
        data[release] = {}
        for arch in release_arches[release]:
            data[release][arch] = {}

    # number of packages with tests
    for (release, arch, numpkgs) in db_con.execute(
            'SELECT release, arch, COUNT(DISTINCT package) '
            'FROM test '
            'GROUP BY release, arch'):
        if numpkgs > 1:
            try:
                data[release][arch]['numpkgs'] = numpkgs
            except KeyError:
                pass

    # number of passed/failed test runs
    for (release, arch, exitcode, numruns) in db_con.execute(
            'SELECT release, arch, exitcode, COUNT(run_id) '
            'FROM test, result '
            'WHERE test.id == result.test_id '
            'GROUP BY release, arch, exitcode'):
        key = 'failruns'
        if exitcode in (0, 2, 8):
            key = 'passruns'
        try:
            data[release][arch][key] = data[release][arch].get(key, 0) + numruns
        except KeyError:
            pass

    # number of packages with tests that pass
    for release in release_arches:
        sources = get_source_versions(db_con, release)
        for arch in release_arches[release]:
            data[release][arch]['numpkgspass'] = success_count_for_release_and_arch(db_con, release, arch, sources)

    return render('browse-statistics.html',
                  release_arches=release_arches,
                  data=data)


if __name__ == '__main__':
    app.config['DEBUG'] = True
    init_config()
    CGIHandler().run(app)
