"""Logic for verifying and submitting test requests

Author: Martin Pitt <martin.pitt@ubuntu.com>
"""

import os
import json
import re
import base64
import logging
import configparser
import sqlite3
import urllib.request
import urllib.parse
from urllib.error import HTTPError
from datetime import datetime

import amqplib.client_0_8 as amqp
from distro_info import UbuntuDistroInfo

# Launchpad REST API base
LP = 'https://api.launchpad.net/1.0/'
NAME = re.compile('^[a-z0-9][a-z0-9.+-]+$')
VERSION = re.compile('^[a-zA-Z0-9.+:~-]+$')
# allowed values are rather conservative, expand if/when needed
ENV = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]+=[a-zA-Z0-9.:~/ -=]*$')
# URL and optional branch name
GIT = re.compile(r'^https?://[a-zA-Z0-9._/~+-]+(#[a-zA-Z0-9._/-]+)?$')

ALLOWED_TEAMS = ['canonical-kernel-distro-team']
# not teams
ALLOWED_USERS_PERPACKAGE = {'snapcraft': ['snappy-m-o']}


class Submit:
    def __init__(self):
        cp = configparser.ConfigParser()
        cp.read(os.path.expanduser('~ubuntu/autopkgtest-cloud.conf'))

        # read valid releases and architectures from DB
        self.db_con = sqlite3.connect('file:%s?mode=ro' % cp['web']['database_ro'], uri=True)
        self.releases = set(UbuntuDistroInfo().supported() + UbuntuDistroInfo().supported_esm())
        logging.debug('Valid releases: %s' % self.releases)

        self.architectures = set()
        c = self.db_con.cursor()
        c.execute('SELECT DISTINCT arch from test')
        while True:
            row = c.fetchone()
            if row is None:
                break
            self.architectures.add(row[0])
        logging.debug('Valid architectures: %s' % self.architectures)

        # dissect AMQP URL
        self.amqp_creds = urllib.parse.urlsplit(cp['amqp']['uri'],
                                                allow_fragments=False)
        assert self.amqp_creds.scheme == 'amqp'
        logging.debug('AMQP credentials: %s' % repr(self.amqp_creds))

    def validate_distro_request(self, release, arch, package, triggers,
                                requester, ppas=[], **kwargs):
        """Validate package and triggers for a distro test request

        'package' is a single source package name. 'triggers' has the format
        ['sourcepackage/version', ...].

        'ppas' is a list of 'team/ppaname' strings.

        Raise ValueError with error messsage if the request is invalid,
        otherwise return.
        """
        can_upload_any_trigger = False

        try:
            if kwargs['delete'] != '1':
                raise ValueError('Invalid delete value')
            del kwargs['delete']
        except KeyError:
            pass
        try:
            if kwargs['all-proposed'] != '1':
                raise ValueError('Invalid all-proposed value')
            del kwargs['all-proposed']
        except KeyError:
            pass
        # no other kwargs supported
        if kwargs:
            raise ValueError('Invalid argument %s' % list(kwargs)[0])

        if release not in self.releases:
            raise ValueError('Unknown release ' + release)
        if arch not in self.architectures:
            raise ValueError('Unknown architecture ' + arch)
        for ppa in ppas:
            if not self.is_valid_ppa(ppa):
                raise ValueError('Unknown PPA ' + ppa)
        # allow kernel tests for EOL vivid
        skip_result_check = (release == 'vivid' and triggers and triggers[0].startswith('linux'))
        if not self.is_valid_package_with_results(None if (ppas or skip_result_check) else release, arch, package):
            raise ValueError('Package %s does not have any test results' %
                             package)
        for trigger in triggers:
            try:
                trigsrc, trigver = trigger.split('/')
            except ValueError:
                raise ValueError('Malformed trigger, must be srcpackage/version')
            # Debian Policy 5.6.1 and 5.6.12
            if not NAME.match(trigsrc) or not VERSION.match(trigver):
                raise ValueError('Malformed trigger')
            if ppas:
                if not self.is_valid_package_version(release, trigsrc, trigver,
                                                     ppas and ppas[-1] or None):
                    raise ValueError('%s is not published in PPA %s %s' %
                                     (trigger, ppas[-1], release))
                # PPAs don't have components, so we need to determine it from the
                # Ubuntu archive
                trigsrc_component = self.is_valid_package_version(release, trigsrc, None) or 'main'
            else:
                trigsrc_component = self.is_valid_package_version(release, trigsrc, trigver)
                if not trigsrc_component:
                    raise ValueError('%s is not published in %s' % (trigger, release))

            can_upload_any_trigger = can_upload_any_trigger or \
                    self.can_upload(requester, release, trigsrc_component, trigsrc)

        if ppas:
            package_component = self.is_valid_package_version(release, package, None) or 'main'
        else:
            package_component = self.is_valid_package_version(release, package, None)
            if not package_component:
                raise ValueError('%s is not published in %s' % (trigger, release))

        # verify that requester can upload package or trigsrc
        if not self.can_upload(requester, release, package_component, package) and \
           not can_upload_any_trigger and \
           requester not in ALLOWED_USERS_PERPACKAGE.get(package, []) and \
           not self.in_allowed_team(requester, package):
            raise ValueError('You are not allowed to upload %s or %s to '
                             'Ubuntu, thus you are not allowed to use this '
                             'service.' % (package, trigsrc))

    def validate_git_request(self, release, arch, package, ppas=[], env=[], **kwargs):
        """Validate parameters for an upstream git test request

        Supported kwargs:
        - 'build-git' is the URL of the branch to test
        - 'ppas' is a list of 'team/ppaname' strings.
        - 'env' is a list of 'key=value' strings.
        - 'testname' is a string.

        Raise ValueError with error messsage if the request is invalid,
        otherwise return.
        """
        if release not in self.releases:
            raise ValueError('Unknown release ' + release)
        if arch not in self.architectures:
            raise ValueError('Unknown architecture ' + arch)
        if not NAME.match(package):
            raise ValueError('Malformed package')
        if not ppas:
            raise ValueError('Must specify at least one PPA (to associate results with)')
        for ppa in ppas:
            if not self.is_valid_ppa(ppa):
                raise ValueError('Unknown PPA ' + ppa)
        for e in env:
            if not ENV.match(e):
                raise ValueError('Invalid environment variable format "%s"' % e)
        # we should only be called in this mode
        assert 'build-git' in kwargs
        if not GIT.match(kwargs['build-git']):
            raise ValueError('Malformed build-git')
        if 'testname' in kwargs and not NAME.match(kwargs['testname']):
            raise ValueError('Malformed testname')

        unsupported_keys = set(kwargs.keys()) - {'build-git', 'testname'}
        if unsupported_keys:
            raise ValueError('Unsupported arguments: %s' % ' '.join(unsupported_keys))

    def unsend_amqp_request(self, release, arch, package, context=None, **params):
        """Remove an autopkgtest AMQP request"""

        if context:
            queue = 'debci-%s-%s-%s' % (context, release, arch)
        else:
            queue = 'debci-%s-%s' % (release, arch)

        count = 0

        with amqp.Connection(self.amqp_creds.hostname,
                             userid=self.amqp_creds.username,
                             password=self.amqp_creds.password) as amqp_con:
            with amqp_con.channel() as ch:
                while True:
                    message = ch.basic_get(queue)
                    if message is None:
                        break
                    if isinstance(message.body, str):
                        body = message.body
                    else:
                        body = message.body.decode('UTF-8')
                    this_package, this_params = body.split(None, 1)
                    this_params = json.loads(this_params)
                    del this_params['submit-time']

                    if this_package == package and this_params == params:
                        ch.basic_ack(message.delivery_tag)
                        count += 1
        return count

    def send_amqp_request(self, release, arch, package, context=None, **params):
        """Send autopkgtest AMQP request"""

        if context:
            queue = 'debci-%s-%s-%s' % (context, release, arch)
        else:
            queue = 'debci-%s-%s' % (release, arch)

        params['submit-time'] = datetime.strftime(datetime.utcnow(), '%Y-%m-%d %H:%M:%S%z')
        body = '%s %s' % (package, json.dumps(params, sort_keys=True))
        with amqp.Connection(self.amqp_creds.hostname,
                             userid=self.amqp_creds.username,
                             password=self.amqp_creds.password) as amqp_con:
            with amqp_con.channel() as ch:
                ch.basic_publish(amqp.Message(body, delivery_mode=2),  # persistent
                                 routing_key=queue)

    @classmethod
    def post_json(klass, url, data, auth_file, project):
        """Send POST request with JSON data via basic auth.

        'data' is a dictionary which will be posted to 'url' in JSON encoded
        form. 'auth_file' is the path to a file containing
        "project:user:password" lines. 'project' is used as a lookup key in
        auth_file.

        This is being used to send status updates to GitHub.

        Raises exception if auth_file does not exist or is invalid. Raises
        HTTPError if POST request fails.
        """
        # look up project in auth_file
        with open(auth_file) as f:
            contents = f.read()
        for l in contents.splitlines():
            if l.startswith(project + ':'):
                credentials = l.split(':', 1)[1].strip()
                break
        else:
            logging.error('%s does not have password for project %s',
                          auth_file, project)
            return

        req = urllib.request.Request(
            url=url, method='POST',
            data=json.dumps(data).encode('UTF-8'),
            headers={'Content-Type': 'application/json',
                     'Authorization': 'Basic %s' % base64.b64encode(credentials.encode()).decode()})
        with urllib.request.urlopen(req) as f:
            logging.debug('POST to URL %s succeeded with code %u: %s',
                          f.geturl(), f.getcode(), f.read())

    #
    # helper methods
    #

    def is_valid_ppa(self, ppa):
        """Check if a ppa exists"""
        team, _, name = ppa.partition('/')
        if not NAME.match(team) or not NAME.match(name):
            return None
        # https://launchpad.net/+apidoc/1.0.html#person-getPPAByName
        (code, response) = self.lp_request('~' + team, {
            'ws.op': 'getPPAByName',
            'distribution': json.dumps(LP + 'ubuntu'),
            'name': json.dumps(name),
        })
        logging.debug(
            'is_valid_ppa(%s): code %u, response %s',
            ppa, code, repr(response))
        if code < 200 or code >= 300:
            return None
        if response.get('name') == name:
            return True

    def is_valid_package_with_results(self, release, arch, package):
        """Check if package exists and has any results on that release+arch

        Use this for validating *tested* packages (not triggers, as they don't
        necessarily have tests themselves).
        """
        c = self.db_con.cursor()
        if release:
            c.execute('SELECT count(arch) FROM test '
                      'WHERE package=? AND arch=? AND release=?',
                      (package, arch, release))
        else:
            c.execute('SELECT count(arch) FROM test '
                      'WHERE package=? AND arch=?',
                      (package, arch))
        return c.fetchone()[0] > 0

    def is_valid_package_version(self, release, package, version, ppa=None):
        """Check if package/version exists in the given release

        Use this for validating trigger packages. This queries the Launchpad
        REST API. Check given ppa (team/name), or the main Ubuntu archive if
        this is None.

        Return the component name if package/version exists, otherwise None.
        """
        # https://launchpad.net/+apidoc/1.0.html#archive-getPublishedSources
        if ppa:
            team, name = ppa.split('/')
            obj = '~%s/+archive/ubuntu/%s' % (team, name)
        else:
            obj = 'ubuntu/+archive/primary'
        req = {'ws.op': 'getPublishedSources',
               'source_name': json.dumps(package),
               'distro_series': json.dumps(LP + 'ubuntu/' + release),
               'status': 'Published',
               'exact_match': 'true'}
        if version is not None:
            req['version'] = json.dumps(version)
        (code, response) = self.lp_request(obj, req)
        if code < 200 or code >= 300:
            return None
        logging.debug(
            'is_valid_package_version(%s, %s, %s): code %u, response %s',
            release, package, version, code, repr(response))
        if response.get('total_size', 0) > 0:
            return response['entries'][0]['component_name']
        else:
            return None

    def can_upload(self, person, release, component, package):
        """Check if person can upload package into Ubuntu release"""

        # https://launchpad.net/+apidoc/1.0.html#archive-checkUpload
        (code, response) = self.lp_request('ubuntu/+archive/primary',
                                           {'ws.op': 'checkUpload',
                                            'distroseries': json.dumps(LP + 'ubuntu/' + release),
                                            'person': json.dumps(LP + '~' + person),
                                            'component': component,
                                            'pocket': 'Proposed',
                                            'sourcepackagename': json.dumps(package),
                                            })
        logging.debug('can_upload(%s, %s, %s, %s): (%u, %s)',
                      person, release, component, package, code, repr(response))
        return code >= 200 and code < 300

    def in_allowed_team(self, person, package=[]):
        """Check if person is in ALLOWED_TEAMS"""

        for team in ALLOWED_TEAMS:
            (code, response) = self.lp_request('~%s/participants' % team, {})
            for e in response.get('entries', []):
                if e.get('name') == person:
                    return True
        return False

    @classmethod
    def lp_request(klass, obj, query):
        """Do a Launchpad REST request

        Request https://api.launchpad.net/1.0/<obj>?<query>.

        Return (code, json), where json is defined for successful codes
        (200 <= code < 300) and None otherwise.
        """
        url = LP + obj + '?' + urllib.parse.urlencode(query)
        try:
            with urllib.request.urlopen(url, timeout=10) as req:
                code = req.getcode()
                if code >= 300:
                    logging.error('URL %s failed with code %u', req.geturl(), code)
                    return (code, None)
                response = req.read()
        except HTTPError as e:
            logging.error('%s failed with %u: %s\n%s', url, e.code, e.reason, e.headers)
            return (e.code, None)

        try:
            response = json.loads(response.decode('UTF-8'))
        except (UnicodeDecodeError, ValueError) as e:
            logging.error('URL %s gave invalid response %s: %s',
                          req.geturl(), response, str(e))
            return (500, None)
        logging.debug('lp_request %s succeeded: %s', url, response)
        return (code, response)
